import sys
import io
import time
import os
import signal
from .util import (
    log_info,
    log_error,
    # log_warning,
    # log_debug,
    log_job_trace,
    shorten_job_id,
)
import select
import termios
import tty
import threading
from .util import console
import rich.status
from collections import namedtuple

StatusReport = namedtuple(
    "StatusReport", ["running", "waiting", "done", "total", "failed"]
)


class ConsoleInteractive:
    def _set_terminal_raw(self):
        """Set almost all raw settings on the terminal, except for the output meddling
        - if we did that we get weird newlines from rich"""
        try:
            self.old_settings = termios.tcgetattr(sys.stdin)
            fd = sys.stdin.fileno()
            when = termios.TCSAFLUSH
            tty.setraw(fd)
            mode = termios.tcgetattr(sys.stdin.fileno())
            mode[1] = mode[1] | termios.OPOST  # termios.tcgetattr(fd)
            termios.tcsetattr(sys.stdin.fileno(), when, mode)
        except io.UnsupportedOperation:  # happens in tests that set it to console mode
            pass

    def _end_terminal_raw(self):
        if hasattr(self, "old_settings"):
            try:
                termios.tcsetattr(sys.stdin, termios.TCSADRAIN, self.old_settings)
            except io.UnsupportedOperation:  # see _set_terminal_raw
                pass

    def start(self, runner):
        self.runner = runner
        self.last_report_status_args = StatusReport(0, 0, 0, len(runner.jobs), 0)
        self.breaker = os.pipe()
        self.thread = threading.Thread(target=self.loop)
        # self._set_terminal_raw()
        self.stopped = False
        self.leave_thread = False
        self.thread.start()
        log_info("Type 'help<enter>' to receive a list of valid commands")
        self._cmd = ""
        self.status = rich.status.Status("", console=console)
        self.status.start()
        self.report_status(self.last_report_status_args)

    def stop(self):
        """Called from the runner"""
        log_info("Leaving interactive mode")
        self.stopped = True
        if hasattr(self, "thread"):
            self.leave_thread = True
            # async_raise(self.thread.ident, KeyboardInterrupt)
            os.write(self.breaker[1], b"x")
            self._end_terminal_raw()
            log_job_trace("Terminating interactive thread")
            self.thread.join()
            log_job_trace("Terminated interactive thread")
            self.status.stop()
        del self.runner
        log_info("Left interactive mode")

    @property
    def cmd(self):
        return self._cmd

    @cmd.setter
    def cmd(self, value):
        self._cmd = value
        self.report_status(self.last_report_status_args)

    def loop(self):
        log_info("Entering interactive loop")
        while True:
            try:
                if self.leave_thread:
                    break
                try:
                    input = select.select([sys.stdin, self.breaker[0]], [], [], 10)[0]
                except io.UnsupportedOperation as e:
                    if "redirected stdin is pseudofile" in str(
                        e
                    ):  # running under pytest - no interactivity, but suport requesting it?
                        input = False
                    else:
                        raise
                if input:
                    if self.breaker[0] in input:
                        break
                    else:  # must have been stdin.
                        value = sys.stdin.read(1)
                        # log_info(f"received {repr(value)}")
                        if value == "\x03":  # ctrl-c:
                            self.cmd = ""
                        elif value == "\x1a":  # ctrl-z
                            os.kill(os.getpid(), signal.SIGTSTP)
                        elif ord("0") <= ord(value) <= ord("z") or value == " ":
                            self.cmd += value
                        elif value == "\x7f":  # backspace
                            self.cmd = self.cmd[:-1]
                        elif value == "\n" or value == "\r":
                            try:
                                if self.cmd:
                                    command = self.cmd
                                    args = ""
                                    if " " in command:
                                        command = command[: command.find(" ")]
                                        args = self.cmd[len(command) + 1 :].strip()
                                    self.cmd = ""
                                    if hasattr(self, "_cmd_" + command):
                                        getattr(self, "_cmd_" + command)(args)
                                    else:
                                        print("No such command")
                                else:
                                    self._cmd_default()
                                self.report_status(self.last_report_status_args)
                            except Exception as e:
                                log_error(
                                    f"An exception occured during command: {e} {type(e)}"
                                )
                                self.cmd = ""
                                continue

            except KeyboardInterrupt:
                break
        # log_job_trace("Leaving interactive loop")

    def report_status(self, report):
        self.last_report_status_args = report
        # msg = f"[dim]Running/Waiting Done/Total[/dim] {report.running} / {report.waiting} {report.done} / {report.total}."  # In flight: {len(self.runner.jobs_in_flight)} "
        msg = f"[dim]T:[/dim]{report.total} D:{report.done} R:{report.running} W:{report.waiting} F:{report.failed}"
        if self.cmd:
            msg += f" Cmd: {self.cmd}"
        else:
            if self.stopped:
                msg += " Exiting..."
            else:
                # msg += "Type help<enter> for commands"
                pass
        self.status.update(status=msg + "\r\n")

    def _cmd_help(self, _args):
        """print help"""
        print("Help for interactive mode")
        print("You have the following commands available")
        print("\t- <enter> - Show currently running jobs")
        for x in dir(self):
            if x.startswith("_cmd"):
                cmd = x[5:]
                if cmd:
                    print(f"\t {cmd} -  {getattr(self, x).__doc__}")
        # print("\t- help - this command")
        # print("\t- abort - kill current jobs and exit asap")
        # print("\t- stop - Wait for the currently running jobs to finish, then exit")
        # print("\t- reboot - After the pipegraph has ended, restart the current python script")
        # print("\t- restart - After the currently running jobs have ended, restart the current python script")

    def _cmd_default(self):
        """print the currently running jobs (mapped to enter)"""
        t = time.time()
        to_sort = []
        for job_id in self.runner.jobs_in_flight:
            try:
                rt = t - self.runner.jobs[job_id].start_time
                to_sort.append((rt, job_id))
            except KeyError:
                pass
        to_sort.sort()
        print(" | ".join(("Job_no", "Runtime", "Cores", "Job_id")))
        print(" | ".join(("------", "-------", "-----", "------")))
        for rt, job_id in to_sort:
            job = self.runner.jobs[job_id]
            job_no = job.job_number
            if job.waiting:
                rt = "waiting"
            else:
                rt = f"{rt:>6.2f}s"
            display_job_id = shorten_job_id(job_id)
            cores = job.actual_cores_needed if job.actual_cores_needed != -1 else "?"
            print(f"{job_no:>6} | {rt} | {cores} | {display_job_id}")
        print("")

    def _cmd_abort(self, _args):
        """Kill current jobs and exit (safely) asap"""
        log_info("Run aborted by command. Safely shutting down")
        self.runner.abort()
        self.stopped = True

    def _cmd_die(self, _args):
        """kill the current process without saving history"""
        log_error("Sic semper running processes")
        os.kill(os.getpid(), signal.SIGTERM)
        sys.exit(1)

    def _cmd_stop(self, _args):
        """Exit after current jobs finished"""
        if not self.stopped:
            log_info("Run stopped by command")
            waiting_for = []
            for job_id in self.runner.jobs_in_flight:
                try:
                    if not getattr(self.runner.jobs[job_id], "waiting", False):
                        waiting_for.append(job_id)
                except KeyError:
                    pass

            log_info(f"Having to wait for jobs: {sorted(waiting_for)}")
            self.runner.stop()
            self.stopped = True

    def _cmd_again(self, _args):
        """Restart the current python program after all jobs have completed"""
        log_info("Again command issued")
        self.runner.job_graph.restart_afterwards()

    def _cmd_stop_and_again(self, _args):
        "Stop after current jobs, then restart the current python program"
        # log_info("Stop_and_again command issued")
        # self.runner.stop()
        # self.stopped = True
        # self.runner.job_graph.restart_afterwards()
        self._cmd_stop(_args)
        self._cmd_again(_args)

    def _cmd_kill(self, args):
        """kill a running job (by id)"""
        try:
            job_no = int(args)
        except ValueError:
            print(f"Could not understand job number {repr(args)}- must be an integer")
            return
        for job in self.runner.jobs.values():
            if job.job_number == job_no:
                break
        else:
            print("Could not find job number")
            return
        if not job.job_id in self.runner.jobs_in_flight:
            print("Job not currently in flight - can't kill it")
            return
        if not job.resources.is_external():
            print("Job is not running in an external process - can't kill")
            return
        print("ok, killing job", job.job_id)
        log_info(f"Command kill {job.job_id} ")
        job.kill_if_running()

    def _cmd_debug(self, args):
        """Write the debug status from teh evaluator to debug.txt"""
        from pathlib import Path

        # import pypipegraph2

        # pypipegraph2.pypipegraph2.enable_logging()
        filename = Path("debug.txt").absolute()
        print(f"Writing to {filename}")
        with open(filename, "w") as op:
            op.write(self.runner.evaluator.debug())
        print(f"Debug information written to {filename}")



import os
import sys
from loguru import logger
from rich.console import Console
from . import ppg_traceback

console_args = {}
if "pytest" in sys.modules:
    console_args["width"] = 120
console = Console(**console_args)

cpu_count = None


def escape_logging(s):
    return str(s).replace("<", "\\<").replace("{", "{{").replace("}", "}}")


def CPUs():
    """
    Detects the number of CPUs on a system. Cribbed from pp.
    """
    global cpu_count
    if cpu_count is None:
        cpu_count = 1  # default
        # Linux, Unix and MacOS:
        if hasattr(os, "sysconf"):
            if "SC_NPROCESSORS_ONLN" in os.sysconf_names:
                # Linux & Unix:
                ncpus = os.sysconf("SC_NPROCESSORS_ONLN")
                if isinstance(ncpus, int) and ncpus > 0:
                    cpu_count = ncpus
            else:  # OSX: pragma: no cover
                cpu_count = int(
                    os.popen2("sysctl -n hw.ncpu")[1].read()
                )  # pragma: no cover
        # Windows:
        if "NUMBER_OF_PROCESSORS" in os.environ:  # pragma: no cover
            ncpus = int(os.environ["NUMBER_OF_PROCESSORS"])
            if ncpus > 0:
                cpu_count = ncpus
    return cpu_count


def job_or_filename(job_or_filename, invariant_class=None):
    """Take a filename, or a job. Return Path(filename), dependency-for-that-file
    ie. either the job, or a invariant_class (default: FileInvariant)"""
    from .jobs import Job, FileInvariant
    from pathlib import Path

    if invariant_class is None:  # pragma: no cover
        invariant_class = FileInvariant

    if isinstance(job_or_filename, Job):
        filename = job_or_filename.files[0]
        deps = [job_or_filename]
    elif job_or_filename is not None:
        filename = Path(job_or_filename)
        deps = [invariant_class(filename)]
    else:
        filename = None
        deps = []
    return filename, deps


def assert_uniqueness_of_object(
    object_with_name_attribute, pipegraph=None, also_check=None
):
    """Makes certain there is only one object with this class & .name.

    This is necessary so the pipegraph jobs assign their data only to the
    objects you're actually working with."""
    if pipegraph is None:  # pragma: no branch
        from pypipegraph2 import global_pipegraph

        pipegraph = global_pipegraph

    if object_with_name_attribute.name.find("/") != -1:
        raise ValueError(
            "Names must not contain /, it confuses the directory calculations"
        )
    if pipegraph is None:  # pragma: no cover
        return
    if not hasattr(pipegraph, "object_uniquifier"):
        pipegraph.object_uniquifier = {}
    typ = object_with_name_attribute.__class__
    if typ not in pipegraph.object_uniquifier:
        pipegraph.object_uniquifier[typ] = {}
    if object_with_name_attribute.name in pipegraph.object_uniquifier[typ]:
        raise ValueError(
            "Doublicate object: %s, %s" % (typ, object_with_name_attribute.name)
        )
    if also_check:
        if not isinstance(also_check, list):
            also_check = [also_check]
        for other_typ in also_check:
            if (
                other_typ in pipegraph.object_uniquifier
                and object_with_name_attribute.name
                in pipegraph.object_uniquifier[other_typ]
            ):
                raise ValueError(
                    "Doublicate object: %s, %s"
                    % (other_typ, object_with_name_attribute.name)
                )
    object_with_name_attribute.unique_id = len(pipegraph.object_uniquifier[typ])
    pipegraph.object_uniquifier[typ][object_with_name_attribute.name] = True


def flatten_jobs(j):
    """Take an arbitrary deeply nested list of lists of jobs
    and return just the jobs"""
    from .jobs import Job

    if isinstance(j, Job):
        yield j
    else:
        for sj in j:
            yield from flatten_jobs(sj)


do_jobtrace_log = False


def log_warning(msg):
    logger.opt(depth=1).warning(msg)


def log_error(msg):
    logger.opt(depth=1).error(msg)


def log_info(msg):
    logger.opt(depth=1).info(msg)


def log_debug(msg):
    logger.opt(depth=1).debug(msg)


def log_job_trace(msg):
    if do_jobtrace_log:
        logger.opt(depth=1).log("JT", msg)


def log_trace(msg):
    if do_jobtrace_log:  # pragma: no cover
        logger.opt(depth=1).trace(msg)


def shorten_job_id(job_id):
    dotdotcount = job_id.count(":::")
    if dotdotcount:
        return job_id[: job_id.find(":::") + 3] + "+" + str(dotdotcount)
    else:
        return job_id


def pretty_log_errors(func):
    """capture exceptions (on a function outside of ppg)
    and format it with our fancy local logging exception logger
    """

    def inner(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as e:
            exception_type, exception_value, tb = sys.exc_info()
            captured_tb = ppg_traceback.Trace(exception_type, exception_value, tb)
            logger.error(
                captured_tb._format_rich_traceback_fallback(
                    include_locals=True, include_formating=True
                )
            )
            raise

    return inner



import networkx
from pathlib import Path
import sys
import subprocess


def browse_graph_ml(filename=None):
    graph = networkx.readwrite.graphml.read_graphml(filename)

    header = "choose a node by typing. Ctrl-c to quit"
    p = subprocess.Popen(
        ["fzf", "--header", header], stdin=subprocess.PIPE, stdout=subprocess.PIPE
    )
    chosen = p.communicate("\n".join(graph.nodes).encode("utf-8"))[0].decode("utf-8")
    stack = [chosen]

    if chosen:
        while True:
            if chosen.startswith("parent: ") or chosen.startswith("child: "):
                chosen = chosen[chosen.find(" ") + 1 :]
            chosen = chosen.strip()
            lines = []
            for child in reversed(sorted(graph.successors(chosen))):
                lines.append(f"child: {child}")
            lines.append("")
            for parent in reversed(sorted(graph.predecessors(chosen))):
                lines.append(f"parent: {parent}")
            lines = "\n".join(lines)
            header = f"Examining: {chosen}"
            p = subprocess.Popen(
                ["fzf", "--header", header],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            chosen = p.communicate(lines.encode("utf-8"))[0].decode("utf-8")
            if not chosen:
                if len(stack) > 1:
                    chosen = stack.pop()
                    chosen = stack.pop()
                else:
                    break
            else:
                if not stack or stack[-1] != chosen:
                    stack.append(chosen)


def cli_browse_graph_ml():
    try:
        filename = sys.argv[1]
    except IndexError:
        print("ppg2_browse_graph_ml <path_to_.graphml>")
        sys.exit(1)
    browse_graph_ml(filename)


def record_as_generated_file(filename, history):
    import pypipegraph2 as ppg2

    if not filename.exists():
        raise FileNotFoundError(filename)

    present = filename in history
    print("already present?", present)
    if present:
        had_input = bool(history[filename][0])
        had_output = bool(history[filename][1])
        print("had output", had_output)
        print("had input", had_input)
    if not present or (not had_output) and (not had_input):
        print("ok, fixing,", filename)
        history[filename] = (
            {},
            {str(filename): ppg2.hashers.hash_file(filename)},
        )
    else:
        print("skipping", filename)


def cli_record_as_generated_file():
    """Record <filename> (relative path) as if it was
    done by a filegenerating job - provided it wasn't recorded
    previously, and it had no input.

    This should not be necessary in almost all circumstances.

    """
    import pypipegraph2 as ppg2

    if len(sys.argv) <= 1:
        print("ppg2_record_as_generated_file <filename>")
        print("Records this file as if it had been generated by the pipegraph")
        print("(only useful if the job in question is not being invalidated")
        sys.exit(1)
    else:
        g = ppg2.new()
        history = g._load_history()

        for filename in sys.argv[1:]:
            record_as_generated_file(Path(filename), history)
        g._save_history(history)



import sys
from loguru import logger  # noqa:F401

sys.path.append("../../python")
import pypipegraph2 as ppg  # noqa:E402
from pathlib import Path  # noqa:E402
import shutil  # noqa:E402


whitelist = ["simplest.py"]
print("\n" * 10)

for fn in Path(".").glob("*"):
    if not fn.name in whitelist:
        if fn.is_dir():
            shutil.rmtree(fn)
        else:
            fn.unlink()

ppg.enable_rust_logging()

ppg.new(log_level=1, cores=2)
ppg.util.do_jobtrace_log = True


class Dummy(object):
    pass


def write(filename, text):
    Path(filename).write_text(text)


def append(filename, text):
    p = Path(filename)
    if p.exists():
        old = p.read_text()
    else:
        old = ""
    p.write_text(old + text)


def writeappend(filename_write, filename_append, string):
    write(filename_write, string)
    append(filename_append, string)


def read(filename):
    return Path(filename).read_text()


def counter(filename):
    """Helper for counting invocations in a side-effect file"""
    try:
        res = int(Path(filename).read_text())
    except:  # noqa: E722
        res = 0
    Path(filename).write_text(str(res + 1))
    return str(res)


def force_load(job):
    """Force the loading of a Dataloading job that has no other dependents"""
    import pypipegraph2 as ppg

    ppg.JobGeneratingJob(job.job_id + "_gen", lambda: None).depends_on(job)


def simplest():
    ppg.FileGeneratingJob("deleteme", lambda of: of.write_text("hello"))


def cached_dl():
    Path("out").mkdir(exist_ok=True)
    o = Dummy()

    def calc():
        return ", ".join(str(x) for x in range(0, 100))

    def store(value):
        o.a = value

    job, cache_job = ppg.CachedDataLoadingJob("out/mycalc", calc, store)
    of = "out/A"

    def do_write(of):
        write(of, o.a)

    ppg.FileGeneratingJob(of, do_write).depends_on(job)
    ppg.run()
    assert read(of) == ", ".join(str(x) for x in range(0, 100))


def test_tempfile():
    jobA = ppg.TempFileGeneratingJob(
        "TA",
        lambda of: of.write_text("A" + counter("a")),
        depend_on_function=False,
    )
    jobB = ppg.FileGeneratingJob(
        "B",
        lambda of: of.write_text("B" + counter("c") + Path("TA").read_text()),
        depend_on_function=False,
    )
    jobB.depends_on(jobA)
    ppg.run()
    assert not Path("TA").exists()
    assert Path("B").exists()
    assert Path("B").read_text() == "B0A0"
    logger.error("Second run - no rerun")
    ppg.run()
    assert not Path("TA").exists()
    assert Path("B").exists()
    assert Path("B").read_text() == "B0A0"

    Path("B").unlink()
    logger.error("Third run - B output missing")
    ppg.run()
    assert not Path("TA").exists()
    assert Path("B").exists()
    assert Path("B").read_text() == "B1A1"


def test_just_chained_tempfile(self, trace_log):
    jobA = ppg.TempFileGeneratingJob("TA", lambda of: of.write_text("A" + counter("a")))
    jobB = ppg.TempFileGeneratingJob(
        "B", lambda of: of.write_text("B" + counter("b") + Path("TA").read_text())
    )
    jobB.depends_on(jobA)
    ppg.run()
    assert not Path("TA").exists()
    assert not Path(
        "a"
    ).exists()  # changed with the smarter hull stuff - they don't run for sideeffects
    assert not Path("B").exists()
    assert not Path("b").exists()


def test_changing_inputs_when_job_was_temporarily_missing(self):
    jobA = ppg.FileGeneratingJob("A", lambda of: counter("a") and of.write_text("AAA"))
    jobB = ppg.FileGeneratingJob(
        "B", lambda of: of.write_text("BBB" + Path("A").read_text())
    )
    jobB.depends_on(jobA)
    ppg.run()
    assert Path("A").read_text() == "AAA"
    assert Path("B").read_text() == "BBBAAA"
    assert Path("a").read_text() == "1"
    ppg.new()
    assert not "A" in ppg.global_pipegraph.jobs
    assert not "B" in ppg.global_pipegraph.jobs
    jobA = ppg.FileGeneratingJob("A", lambda of: counter("a") and of.write_text("AAAA"))
    assert not "B" in ppg.global_pipegraph.jobs
    ppg.run()
    assert Path("A").read_text() == "AAAA"
    assert Path("B").read_text() == "BBBAAA"  # not rerun
    assert Path("a").read_text() == "2"
    ppg.new()
    jobA = ppg.FileGeneratingJob("A", lambda of: counter("a") and of.write_text("AAAA"))
    jobB = ppg.FileGeneratingJob(
        "B", lambda of: of.write_text("BBB" + Path("A").read_text())
    )
    jobB.depends_on(jobA)
    ppg.run()
    assert Path("a").read_text() == "2"
    assert Path("B").read_text() == "BBBAAAA"


# test_changing_inputs_when_job_was_temporarily_missing(None)


def test_tempfile_chained_invalidate_leaf(self):
    ppg.new(cores=1, log_level=6)
    jobA = ppg.TempFileGeneratingJob(
        "TA", lambda of: of.write_text("A" + counter("a")), depend_on_function=False
    )
    jobB = ppg.TempFileGeneratingJob(
        "TB",
        lambda of: of.write_text("B" + counter("b") + Path("TA").read_text()),
        depend_on_function=False,
    )
    jobC = ppg.FileGeneratingJob(
        "C",
        lambda of: of.write_text("C" + counter("c") + Path("TB").read_text()),
        depend_on_function=False,
    )
    jobC.depends_on(jobB)
    jobB.depends_on(jobA)
    ppg.util.log_error("First run")
    ppg.run()
    assert not Path("TA").exists()
    assert not Path("TB").exists()
    assert Path("C").read_text() == "C0B0A0"
    logger.error("Second No op run.")
    ppg.run()
    assert Path("C").read_text() == "C0B0A0"
    assert not Path("TA").exists()
    assert not Path("TB").exists()

    jobC.depends_on(ppg.FunctionInvariant(lambda: 53, "lambda_52"))
    ppg.util.log_error("Third run - rerun because of FI")
    ppg.run()
    assert Path("C").read_text() == "C1B1A1"
    assert not Path("TA").exists()
    assert not Path("TB").exists()


test_tempfile_chained_invalidate_leaf(None)

ppg.run()



# check if we can abort long running & stalled jobs
import pypipegraph2 as ppg
import time
from pathlib import Path
import logging
import os
import shutil
import psutil
import subprocess

p = Path("run/stall")
if p.exists():
    shutil.rmtree(p)
p.mkdir(exist_ok=True, parents=True)
os.chdir(p)

time_before_abort = 1
timeout = 60
jobcount = 5
ppg.new(log_level=logging.INFO, cores=5)
# tell the ppg to accept the SIGINT we're actually sending
ppg.global_pipegraph._debug_allow_ctrl_c = "abort"


def all_cores(ii):
    def inner(of, ii=ii):
        of.write_text(str(time.time()))
        proc = psutil.Process()
        parent = proc.parent()
        if ii == 0:  # only the first guy kills us
            time.sleep(time_before_abort)
            for tt in range(timeout):
                subprocess.check_call(["kill", "--signal", "SIGINT", str(parent.pid)])
                time.sleep(1)

    return ppg.FileGeneratingJob(
        f"all_cores{ii}",
        inner,
        resources=ppg.Resources.AllCores,
        depend_on_function=False,
    )


ac = [all_cores(ii) for ii in range(jobcount)]

start = time.time()
try:
    ppg.run()
except KeyboardInterrupt:
    print("Received expected Keyboard interrupt")
    stop = time.time()
    print(
        f"Abort took {stop-start-time_before_abort:.2f} seconds.\n "
        "If jobs were completed, you'd have observed something closer to {timeout}"
    )



import json
from .util import (
    # log_info,
    log_error,
    # log_warning, log_debug, log_trace,
    # log_job_trace
)
import sys
from . import ppg_traceback


def history_is_different(runner, job_upstream_id, job_downstream_id, str_last, str_now):
    # note that at this point, we already know str_last != str_now,
    # that was tested in rust
    # log_error(f"history is maybe different {job_upstream_id} {job_downstream_id} {str_last == str_now}")
    job_upstream = runner.jobs[job_upstream_id]
    obj_last = json.loads(str_last)
    obj_now = json.loads(str_now)
    if job_downstream_id == "!!!":
        # special case where we compare a job to itself, not to the input it delivered into another job.
        # Has to do with ephemeral jobs not changing output when validated-but-rerun.
        outputs = job_upstream.outputs
        for ip in outputs:
            altered = not job_upstream.compare_hashes(obj_last[ip], obj_now[ip])
            if altered:
                log_error(f"history is actually different {obj_last[ip]} {obj_now[ip]}")
                return True

    else:
        try:
            outputs = job_upstream.outputs
            inputs = runner.job_inputs[job_downstream_id]
            for ip in inputs:
                if ip in outputs:
                    altered = not job_upstream.compare_hashes(obj_last[ip], obj_now[ip])
                    if altered:
                        log_error(
                            f"history is actually different {obj_last[ip]} {obj_now[ip]}"
                        )
                        return True
        except:  # noqa: E722 yes we really want to capture and reraise *everything*
            exception_type, exception_value, tb = sys.exc_info()
            captured_tb = ppg_traceback.Trace(exception_type, exception_value, tb)
            log_error(f"old was {str_last}")
            log_error(f"now was {str_now}")
            log_error(f"{captured_tb}")
            raise

    # log_error(f"history the same {job_upstream_id} {job_downstream_id}")
    return False



import subprocess
import sys
import os
from pathlib import Path
import psutil

import setproctitle

setproctitle.setproctitle("stage1")


p = subprocess.Popen(["python", Path(__file__).parent / "stage2.py"], env=os.environ)
p.communicate()
print("stage 2 ended, ", os.getpid())
subprocess.check_call(["ps", "--forest", "-o", "%p %r %c"])

for proc in psutil.process_iter(["cmdline"]):
    if "ppg2_test_parallel_sentinel" in proc.info["cmdline"]:
        print(
            "found a ppg2_test_parallel_sentinel", "pid", proc.pid, "ppid", proc.ppid()
        )
        sys.exit(1)
    if "sleep" in proc.info["cmdline"]:
        print("found a sleep", "pid", proc.pid, "ppid", proc.ppid())
        print(proc.pid, proc.ppid())
        sys.exit(1)

sys.exit(0)



import pytest
import pypipegraph2 as ppg


@pytest.mark.usefixtures("ppg2_per_test")
class TestUtils:
    def test_assert_uniqueness_simple(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        Dummy("shu")

        with pytest.raises(ValueError):
            Dummy("shu")

    def test_assert_uniqueness_ok(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        Dummy("shu")
        Dummy("sha")

        with pytest.raises(ValueError):
            Dummy("shu")

    def test_assert_uniqueness_ok_multi_classes(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        class Dummy2:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        Dummy("shu")
        Dummy2("shu")

        with pytest.raises(ValueError):
            Dummy("shu")

    def test_assert_uniqueness_raises_slashes(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        Dummy("shu")

        with pytest.raises(ValueError):
            Dummy("shu/sha")

    def test_assert_uniqueness_raises_also_check(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        class Dummy2:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self, also_check=Dummy)

        Dummy("shu")

        with pytest.raises(ValueError):
            Dummy2("shu")

    def test_assert_uniqueness_raises_also_check_no_instance_of_second_class(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        class Dummy2:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self, also_check=Dummy)

        # a = Dummy('shu')
        # does not raise of course...
        Dummy2("shu")

        with pytest.raises(ValueError):
            Dummy2("shu")

    def test_assert_uniqueness_raises_also_check_list(self):
        class Dummy:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self)

        class Dummy2:
            def __init__(self, name):
                self.name = name
                ppg.util.assert_uniqueness_of_object(self, also_check=[Dummy])

        Dummy("shu")

        with pytest.raises(ValueError):
            Dummy2("shu")

    def test_exception_on_run_without_previous_new_pipegraph(self):
        ppg.global_pipegraph = None
        with pytest.raises(ValueError):
            ppg.run()

    def test_flatten_jobs(self):
        ppg.new(run_mode=ppg.RunMode.NOTEBOOK)
        j1 = ppg.FileGeneratingJob("A", lambda of: "A")
        j2 = ppg.FileGeneratingJob("B", lambda of: "B")
        j3 = ppg.FileGeneratingJob("B", lambda of: "C")
        res = [j1, [j2, [j3, j1]]]
        # no dedup on this.
        assert list(ppg.util.flatten_jobs(res)) == [j1, j2, j3, j1]

    def test_inside_ppg(self):
        assert ppg.global_pipegraph is not None
        assert ppg.inside_ppg()
        ppg.global_pipegraph = None
        assert not ppg.inside_ppg()



import multiprocessing
from pathlib import Path
from . import exceptions
import sys
import os
import json
import queue
import time
import networkx
from .util import escape_logging
from .enums import (
    JobOutcome,
    RunMode,
)
from .exceptions import _RunAgain
from .parallel import CoreLock, async_raise, FakeLock
from threading import Thread
from . import ppg_traceback
import threading
from .interactive import ConsoleInteractive, StatusReport
from .util import (
    log_info,
    log_error,
    # log_warning,
    # log_debug,
    log_trace,
    log_job_trace,
)
import copy
from .job_status import RecordedJobOutcome
from .history_comparisons import history_is_different
from collections import deque
import signal
import psutil
import ctypes


watcher_parent_pid = None
watcher_ignored_processes = None
watcher_session_id = None


def get_same_session_id_processes():
    for proc in psutil.process_iter():
        proc_sid = os.getpgid(proc.pid)
        if proc_sid == watcher_session_id:
            yield proc


def kill_process_group_but_ppg_runner():
    log_info("kill_process_group_but_ppg_runner")
    children_to_reap = []
    my_pid = os.getpid()
    for proc in get_same_session_id_processes():
        log_info(f"{proc} ignored: {proc in watcher_ignored_processes}")
        if not proc in watcher_ignored_processes:
            children_to_reap.append(proc)
    for p in children_to_reap:
        try:
            log_info(
                f"Abort: Watcher is terminating child processes {p.pid} {p.name()} {p.status()}"
            )
            p.terminate()
        except psutil.NoSuchProcess:
            pass
    gone, alive = psutil.wait_procs(children_to_reap, timeout=2)
    for p in alive:
        log_info(
            f"Abort: Watcher is killing child processes {p.pid} {p.name()} {p.status()}"
        )
        p.kill()
    gone, alive = psutil.wait_procs(children_to_reap, timeout=2)
    if alive:
        log_info(f"Still alive? {alive}")


def watcher_unexpected_death_of_parent(_signum, _frame):
    log_error(f"Watcher discovered parent process died unexpectedly {os.getpid()}")
    kill_process_group_but_ppg_runner()
    os._exit(0)


def watcher_expected_death_of_parent(_signum, _frame):
    log_info("Watcher discovered parent process ended regularly")
    kill_process_group_but_ppg_runner()
    os._exit(0)


def spawn_watcher():
    """The watcher registers to be informed when the parent process dies.
    It then goes and kills the process group,
    which we use to reliably tag the spawned processes.

    That means you can't spawn a daemon process from a ppg job.

    The watcher also get's a sig_int when the graph execution ends,
    which also signals it to go and terminate all stragglers.

    """
    global watcher_parent_pid, watcher_ignored_processes, watcher_session_id
    if watcher_session_id is None:
        log_info("Capturing watcher process group id")
        my_pid = os.getpid()
        my_pgid = os.getpgid(my_pid)
        # ok, the shell gives us a group id, right?
        # we do not mess with that, or the session id, we are not a shell,
        # and we don't fork for the top level managment process.
        # (messing with setgrp and so on has only lead to pain...)
        watcher_session_id = my_pgid
        log_info(f"Watcher identifies process by process group id {watcher_session_id}")
    recv, send = multiprocessing.Pipe()
    log_info("Collecting already running process that will be ignored by teh watcher")
    log_info("%s" % (watcher_ignored_processes,))
    pid = os.fork()
    if pid == 0:
        watcher_ignored_processes = list(get_same_session_id_processes())
        watcher_parent_pid = os.getppid()
        parent_process = psutil.Process(watcher_parent_pid)
        # these guys were around before, so we don't kill tehm.

        signal.signal(signal.SIGTERM, watcher_unexpected_death_of_parent)
        signal.signal(signal.SIGINT, watcher_expected_death_of_parent)
        # get watcher to detect parent's death
        prtcl = ctypes.CDLL("libc.so.6")["prctl"]
        PR_SET_PDEATHSIG = 1
        prtcl(PR_SET_PDEATHSIG, signal.SIGTERM)  # 1 =
        log_info("entering watcher loop")
        send.send_bytes(b"go")
        send.close()
        sys.stdin.close()
        while True:
            time.sleep(100)  # we Live by signal from here on
        log_info("watcher loop exit - should be unreachable?")
        os._exit(0)
    else:
        recv.recv_bytes()
        recv.close()
        return pid


ljt = log_job_trace


ExitNow = "___!!!ExitNow!!___"
#   """Token for leave-this-thread-now-signal"""


class Runner:
    """Run a given JobGraph"""

    def __init__(
        self,
        job_graph,
        history,
        event_timeout,
        focus_on_these_jobs,
        jobs_already_run_previously,
        dump_graphml,
        run_id,
        jobs_do_dump_subgraph_debug,
    ):
        from . import _with_changed_global_pipegraph

        log_trace("Runner.__init__")
        self.event_timeout = event_timeout
        with _with_changed_global_pipegraph(JobCollector(job_graph.run_mode)):
            self.job_graph = job_graph
            self.jobs = job_graph.jobs.copy()
            self.job_inputs = copy.deepcopy(
                job_graph.job_inputs
            )  # job_graph.job_inputs.copy()
            self.outputs_to_job_ids = job_graph.outputs_to_job_ids.copy()
            self.next_job_number = self.job_graph.next_job_number
            self.core_lock = CoreLock(job_graph.cores)
            self.fail_counter = 0
            self.run_id = (
                run_id  # to allow jobgenerating jobs to run just once per graph.run()
            )

            if not networkx.algorithms.is_directed_acyclic_graph(
                self.job_graph.job_dag
            ):  # pragma: no cover - defensive
                error_fn = self.job_graph.log_dir / "debug_edges_with_cycles.txt"
                networkx.write_edgelist(self.job_graph.job_dag, error_fn)
                cycles = list(networkx.simple_cycles(self.job_graph.job_dag))
                raise exceptions.NotADag(
                    f"Not a directed *acyclic* graph. See {error_fn}. Cycles between {cycles}"
                )
            assert len(self.jobs) == len(job_graph.job_dag)

            log_job_trace(f"Focus on these jobs: {focus_on_these_jobs}")
            log_job_trace(f"jobs_already_run_previously: {jobs_already_run_previously}")
            self.dag, self.pruned = self.modify_dag(
                job_graph,
                focus_on_these_jobs,
                jobs_already_run_previously,
                history,  # do we need that?
                dump_graphml,
            )

            job_numbers = set()
            for job_id, job in self.jobs.items():
                # log_job_trace(f"{job_id} {type(self.jobs[job_id])}")
                if job.job_number in job_numbers:
                    raise ValueError(
                        "Duplicate job_number", job.job_number, job_id, job
                    )
                job_numbers.add(job.job_number)
            assert len(job_numbers) == len(self.jobs)
            if len(self.jobs) - len(self.pruned) != len(self.dag):
                raise NotImplementedError(
                    f"Mismatch between len(self.jobs) {len(self.jobs)} - prune_counter {len(self.pruned)} and len(self.dag) {len(self.dag)}"
                )

            log_job_trace(
                "dag "
                + escape_logging(
                    json.dumps(
                        networkx.readwrite.json_graph.node_link_data(self.dag), indent=2
                    )
                ),
            )

            if not networkx.algorithms.is_directed_acyclic_graph(
                self.dag
            ):  # pragma: no cover - defensive
                error_fn = self.job_graph.log_dir / "debug_edges_with_cycles.txt"
                networkx.write_edgelist(self.dag, error_fn)
                cycles = list(networkx.simple_cycles(self.dag))
                raise exceptions.NotADag(
                    f"Not a directed *acyclic* graph after modification. See {error_fn}. Cycles between {cycles}"
                )

            self.event_lock = threading.Lock()
            self.jobs_to_run_que = queue.PriorityQueue()
            self.threads = []
            self.jobs_that_need_propagation = deque()
            if jobs_do_dump_subgraph_debug:
                j1 = self.jobs[list(jobs_do_dump_subgraph_debug)[0]]
                j1.dump_subgraph_for_debug(
                    jobs_do_dump_subgraph_debug, self.jobs, self.dag
                )

    def _apply_pruning(self, dag, focus_on_these_jobs, jobs_already_run_previously):
        def _recurse_pruning(job_id, reason):
            """This goes forward/downstream"""
            pruned.add(job_id)
            if not hasattr(self.jobs[job_id], "prune_reason"):
                self.jobs[job_id].prune_reason = reason
            for downstream_job_id in dag.successors(job_id):
                _recurse_pruning(downstream_job_id, reason)

        def _recurse_unpruning(job_id):
            """This goes upstream"""
            # log_job_trace(f"_recurse_unpruning {job_id}")
            try:
                pruned.remove(job_id)
                del self.jobs[job_id].prune_reason
            except (KeyError, AttributeError):
                pass
            for downstream_job_id in dag.predecessors(job_id):
                _recurse_unpruning(downstream_job_id)

        if jobs_already_run_previously:
            new_jobs = set(dag.nodes).difference(jobs_already_run_previously)
            ljt(f"new jobs {new_jobs}")
        else:
            new_jobs = set()

        pruned = set()
        if focus_on_these_jobs:  # which is only set in the first run...
            # prune all jobs,
            # then unprune this one and it's predecessors
            pruned.update(set(dag.nodes))  # prune all...
            focus_job_ids = set((x.job_id for x in focus_on_these_jobs))
            for job_id in focus_job_ids.union(new_jobs):
                _recurse_unpruning(job_id)
        else:
            focus_job_ids = set()

        # apply regular pruning
        if jobs_already_run_previously:
            pruned.update(jobs_already_run_previously)

        for job_id in self.jobs:
            if self.jobs[job_id]._pruned:
                if not job_id in focus_job_ids:
                    log_job_trace(f"pruning because of _pruned {job_id}")
                    _recurse_pruning(job_id, job_id)

        for job_id in new_jobs:
            _recurse_unpruning(job_id)

        for job_id in pruned:
            log_job_trace(f"pruned {job_id}")
            try:
                dag.remove_node(job_id)
            except (
                networkx.exception.NetworkXError
            ):  # happens with cleanup nodes that we  omitted
                pass
            # del self.jobs[job_id]
        return pruned

    def modify_dag(  # noqa: C901
        self,
        job_graph,
        focus_on_these_jobs,
        jobs_already_run_previously,
        history,
        dump_graphml,
    ):
        """Modify the DAG to be executed
        by pruning
            - focusing on selected jobs (i.e. prune everything outside of their connected component)
            - removing jobs we ran in the last run-through

        """

        dag = job_graph.job_dag.copy()
        if dump_graphml:
            for node in dag.nodes():
                dag.nodes[node]["label"] = node
                dag.nodes[node]["shape"] = self.jobs[node].__class__.__name__
            networkx.readwrite.graphml.write_graphml(
                dag,
                self.job_graph.log_dir / "graph_pre_mod.graphml",
                named_key_ids=True,
            )

        pruned = self._apply_pruning(
            dag, focus_on_these_jobs, jobs_already_run_previously
        )
        return dag, pruned

    def compare_history(self, job_id_from, job_id_to, last_value, new_value):
        # called from rust
        return history_is_different(self, job_id_from, job_id_to, last_value, new_value)

    @staticmethod
    def get_job_inputs_str(job_graph, job_id):
        # called from rust...
        inputs = sorted(job_graph.job_inputs[job_id])
        return "\n".join(inputs)

    def build_evaluator(self, history):
        from .pypipegraph2 import PPG2Evaluator

        e = PPG2Evaluator(
            history,
            self.compare_history,
            lambda job_id: self.__class__.get_job_inputs_str(self.job_graph, job_id),
        )
        # todo: see how much we can push into rust of
        # the whole networkx business.
        # no need to keep multiple graphs, I suppose.
        for job_id in self.dag.nodes:
            job = self.jobs[job_id]
            e.add_node(job_id, job.eval_job_kind)
        for a, b in self.dag.edges:
            e.add_edge(b, a)
        return e

    def run(self, history, last_job_states, print_failures):  # noqa:C901
        """Actually run the current DAG"""
        from . import global_pipegraph

        job_count = len(global_pipegraph.jobs)  # track if new jobs are being created

        log_trace("Runner.__run__")

        self.pid = (
            os.getpid()
        )  # so we can detect if we return inside a forked process and exit (safety net)
        self.start_time = time.time()
        self.aborted = False
        self.stopped = False
        self.print_failures = print_failures
        self.output_hashes = {}
        self.last_job_states = last_job_states

        self.history = history  # so the jobs can peak at it and avoid reprocessing

        self.evaluator = self.build_evaluator(history)
        self.evaluator_lock = threading.Lock()
        self.evaluator.event_startup()
        self.evaluation_done = threading.Event()
        self.check_for_new_jobs = threading.Event()

        self.jobs_in_flight = []
        self.jobs_all_cores_in_flight = 0
        self.job_outcomes = {}

        self.watcher_pid = spawn_watcher()
        self._start_job_executing_threads()

        self.jobs_done = 0
        try:
            self._interactive_start()
            while True:
                # ljt("Waiting for evaluation done")
                if self.evaluation_done.wait(5):  # todo: timeout, sanity checks.
                    # ljt(f"evaluation done happend? {self.stopped}")
                    break
                else:  # periodic timeout
                    if self.aborted:
                        break
            # self._interactive_report()

        finally:
            # log_job_trace("Joining threads")
            with self.evaluator_lock:
                self.stopped = True
            for t in self.threads:
                t.join()
            if self.aborted:  # todo: refactor
                log_error("run was aborted")

                #           reap_children(children_before_run)

            if hasattr(self, "_status"):  # todo: what is this? rich status?
                print("status stop necessary")
                self._status.stop()
            # log_info("interactive stop")
            self._interactive_stop()
            # unregister_reaper(reaper)
            assert psutil.pid_exists(self.watcher_pid)
            os.kill(self.watcher_pid, signal.SIGINT)
            gone, alive = psutil.wait_procs(
                [psutil.Process(self.watcher_pid)], timeout=10
            )
            if alive:
                log_error("Watcher was still hanging around?!")
            # os.waitpid(self.watcher_pid, 1)
            log_info("finished waiting for watcher")

        for job_id in self.pruned:
            if job_id in self.job_outcomes:
                raise ValueError("Pruned & having outcome?!")
            self.job_outcomes[job_id] = RecordedJobOutcome(
                job_id, JobOutcome.Pruned, None
            )
            ljt(f"Logging as pruned {job_id}")

        for job_id in self.evaluator.list_upstream_failed_jobs():
            ljt(f"upstream failed {job_id}")
            if job_id in self.job_outcomes:
                raise ValueError("Upstream Failed & having other outcome?!")
            self.job_outcomes[job_id] = RecordedJobOutcome(
                job_id, JobOutcome.UpstreamFailed, None
            )  # todo: payload = which upstream

        for job_id in self.jobs:
            if not job_id in self.job_outcomes:
                self.job_outcomes[job_id] = RecordedJobOutcome(
                    job_id, JobOutcome.Skipped, None
                )

        if len(global_pipegraph.jobs) != job_count and not self.aborted:
            log_info(
                f"created new jobs. _RunAgain issued {len(global_pipegraph.jobs)} != {job_count}"
            )
            for job_id in global_pipegraph.jobs:
                if job_id not in self.jobs:
                    log_job_trace(f"new job {job_id}")
            raise _RunAgain(
                (self.job_outcomes, self.evaluator.new_history())
            )  # self.job_states)
        log_trace("Left runner.run()")
        if self.aborted:
            self.evaluator.event_abort()

        return self.job_outcomes, self.evaluator.new_history()

    def _interactive_start(self):
        """Activate the interactive thread"""
        if self.job_graph.run_mode is RunMode.CONSOLE:
            self.interactive = ConsoleInteractive()
            self.last_status_time = time.time()
            self.interactive.start(self)

    def _interactive_stop(self):
        """Stop the interactive thread (if present)"""
        if hasattr(self, "interactive"):
            self.interactive.stop()

    def _interactive_report(self):
        if hasattr(self, "interactive"):
            t = time.time()
            if (
                t - self.last_status_time >= 0.5
            ):  # don't update more than every half second.
                waiting = len(
                    [
                        x
                        for x in self.jobs_in_flight
                        if getattr(self.jobs[x], "waiting", False)
                    ]
                )
                self.interactive.report_status(
                    StatusReport(
                        len(self.jobs_in_flight) - waiting,
                        waiting,
                        self.jobs_done,
                        len(self.dag),
                        self.fail_counter,
                    )
                )
                self.last_status_time = t

    def abort(self):
        """Kill all running jobs and leave runner.
        Called from the interactive interface
        """
        log_info("runner.abort called")
        self.abort_time = time.time()
        self.aborted = True
        self.check_for_new_jobs.set()
        self.evaluation_done.set()
        for t in self.threads:
            try:
                async_raise(t.ident, KeyboardInterrupt)
            except ValueError:
                pass

    def stop(self):
        """Leave runner after current jobs
        Called from the interactive interface

        """
        self.stopped = True
        self.abort_time = time.time()

    def _start_job_executing_threads(self):
        """Fire up the default number of threads"""
        for ii in range(self.job_graph.cores):
            self._start_another_thread()

    def _job_failed_last_time(self, job_id) -> bool:
        """Did this job fail last time?"""
        res = (
            self.last_job_states
            and job_id in self.last_job_states
            and self.last_job_states[job_id].outcome == JobOutcome.Failed
        )
        log_trace(f"_job_failed_last_time: {job_id}: {res}")
        return res

    def _start_another_thread(self):
        """Fire up another thread (if all current threads are blocked with multi core threads.

        This prevents stalling, since it will ensure that there's a thread around
        to do the SingleCore jobs.

        Note that we don't fire up threads without limit - at one point, you can still
        stall the graph
        """
        if self.stopped or self.aborted:
            return
        t = Thread(target=self._executing_thread)
        self.threads.append(t)
        t.start()

    def log_failed_job(self, job_id, error):
        if self.print_failures:
            log = log_error
        else:
            log = log_job_trace
        if True:  # not self._job_failed_last_time(job_id): # Todo
            try:
                job = self.jobs[job_id]
                # mock failure in case of abort/stop
                if isinstance(error, exceptions.JobCanceled):
                    if self.aborted or self.stopped:
                        return
                    else:
                        raise NotImplementedError(
                            "JobCanceled outside of stopped/aborted state?!"
                        )
                # log error to file. Todo: move to job_state
                if self.job_graph.error_dir is not None:
                    error_file = (
                        self.job_graph.error_dir
                        / self.job_graph.time_str
                        / (str(job.job_number) + "_exception.txt")
                    )
                else:
                    error_file = None

                if hasattr(error.args[1], "stacks"):
                    stacks = error.args[1]
                    log_info(
                        f"\n{job_id}\n\t"
                        + escape_logging(
                            stacks._format_rich_traceback_fallback(False, False)
                        ).replace("\n", "\n\t")
                        + "\n"
                    )
                else:
                    stacks = None
                    log(error)
                    log("no stack available")

                if self.job_graph.error_dir is not None:
                    with open(error_file, "w") as ef:
                        ef.write(f"JobId: {job_id}\n")
                        ef.write(f"Class: {job.__class__.__name__}\n")
                        ef.write("Input jobs:\n")
                        for parent_id in sorted(self.dag.predecessors(job_id)):
                            ef.write(
                                f"\t{parent_id} ({self.jobs[parent_id].__class__.__name__})\n"
                            )
                        ef.write("\n\n")
                        if stacks is not None:
                            ef.write(
                                stacks._format_rich_traceback_fallback(
                                    include_locals=True, include_formating=False
                                )
                            )
                        else:
                            ef.write(str(error))
                            ef.write("no stack available")
                        if hasattr(job, "stdout"):
                            ef.write("\n\n")
                            ef.write("job stdout:\n")
                            ef.write(str(job.stdout))
                        else:
                            ef.write("\n\nstdout: not available\n")
                        if hasattr(job, "stderr"):
                            ef.write("\n\n")
                            ef.write("job stderr:\n")
                            ef.write(str(job.stderr))
                        else:
                            ef.write("\n\nstderr: not available\n")
                        ef.flush()

                    log(
                        f"\n\tMore details (stdout, locals) in {error_file}\n"
                        f"\tFailed after {job.run_time:.2}s.\n"
                        f"\t{job_id}\n"
                    )
                else:
                    log(f"Failed job: {job_id}")

            except Exception as e:
                log_error(
                    f"An exception ocurred reporting on a job failure for {job_id}: {e}. The original job failure has been swallowed."
                )
        else:
            raise ValueError("Did not expect this")

    def _executing_thread(self):
        """The inner function of the threads actually executing the jobs"""
        cwd = (
            os.getcwd()
        )  # so we can detect if the job cahnges the cwd (don't do that!)
        fake_lock = FakeLock()
        try:
            try:
                while not (self.stopped or self.aborted):
                    job_id = None
                    error = None
                    outputs = None
                    try:
                        do_sleep = False
                        with self.evaluator_lock:
                            cleanups = self.evaluator.jobs_ready_for_cleanup()

                            for cleanup_job_id in cleanups:
                                try:
                                    log_error(f"Cleanup for {cleanup_job_id}")
                                    self.jobs[cleanup_job_id].cleanup()
                                except Exception as e:
                                    log_error(f"Cleanup had an exception {repr(e)}")
                                self.evaluator.event_job_cleanup_done(cleanup_job_id)

                            rr = self.evaluator.jobs_ready_to_run()
                            if not rr:
                                if self.evaluator.is_finished():
                                    # ljt("detected finished")
                                    self.stopped = True
                                    self.evaluation_done.set()
                                    break
                                else:
                                    # ljt("not finished")
                                    # this happens when we have more threads than jobs?
                                    # to run at *this* junction.
                                    do_sleep = True
                                    if not self.jobs_in_flight:
                                        log_error(
                                            f"jobs the engine considers running {self.evaluator.jobs_running()}"
                                        )
                                        log_error(
                                            f"The engine considers itself finished = {self.evaluator.is_finished()}"
                                        )
                                        log_error(
                                            "Evaluator is not finished, reports no jobs ready to run, but no jobs currently running -> a bug in the state machine. No way forward, aborting graph executing (cleanly). Graph written to debug.txt "
                                        )
                                        self.evaluator.debug_is_finished()  # if this helps' we're looking at a propagation failure. Somewhere.
                                        Path("debug.txt").write_text(
                                            self.evaluator.debug()
                                        )  # if this helps' we're looking at a propagation failure. Somewhere.
                                        # self.evaluator.reconsider_all_jobs() # if this helps' we're looking at a propagation failure. Somewhere.
                                        # if not self.evaluator.is_finished() and not self.evaluator.jobs_ready_to_run():
                                        #     #print("goin gdown hard")
                                        #     #self.interactive._cmd_die(False)
                                        # else:
                                        #     log_error("Could recover by doing reconsider_all")

                            else:
                                ljt(f"to run {rr}")
                                job_id = rr[0]
                                self.jobs_in_flight.append(job_id)
                                # ljt(f"added {job_id} {self.jobs_in_flight}")
                                self.evaluator.event_now_running(job_id)

                                job = self.jobs[job_id]
                                job.waiting = True
                                job.actual_cores_needed = -1
                                self._interactive_report()
                                job.start_time = (
                                    time.time()
                                )  # assign it just in case anything fails before acquiring the lock
                                job.stop_time = float("nan")
                                job.run_time = float("nan")

                                c = job.resources.to_number(self.core_lock.max_cores)
                                job.actual_cores_needed = c
                                # log_trace(
                                # f"{job_id} cores: {c}, max: {self.core_lock.max_cores}, jobs_in_flight: {len(self.jobs_in_flight)}, all_cores_in_flight: {self.jobs_all_cores_in_flight}, threads: {len(self.threads)}"
                                # )
                                job_failed_last_time = self._job_failed_last_time(
                                    job_id
                                )
                                if not job_failed_last_time and (c > 1):
                                    # we could stall all SingleCores/RunsHere by having all_cores blocking all but one thread (which executes another all_core).
                                    # if we detect that situation, we spawn another one.
                                    self.jobs_all_cores_in_flight += 1
                                    if (
                                        self.jobs_all_cores_in_flight
                                        >= len(self.threads)
                                        and len(self.threads)
                                        <= self.job_graph.cores
                                        * 5  # at one point, we either have to let threads die again, or live with
                                        # the wasted time b y stalling.
                                    ):
                                        ljt(
                                            "All threads blocked by Multi core jobs - starting another one"
                                        )
                                        self._start_another_thread()
                        # letting go of evaluator lock
                        if do_sleep:
                            self.check_for_new_jobs.wait(
                                60
                            )  # not sure why we even have a timeout?
                            self.check_for_new_jobs.clear()
                            continue
                        else:
                            pass

                        # ljt(f"wait for {job_id}")
                        if c == 0:
                            log_error(f"Cores was 0! {job.job_id} {job.resources}")

                        if job_failed_last_time:
                            lock = fake_lock
                        else:
                            lock = self.core_lock

                        with lock.using(c):
                            if self.stopped or self.aborted:
                                continue  # -> while not stopped -> break
                            job.start_time = time.time()  # the *actual* start time
                            job.waiting = False
                            self._interactive_report()
                            ljt(f"Go {job_id}")

                            try:
                                # that's history-output
                                old_history_for_this_job = self.history.get(
                                    job_id, None
                                )
                                if old_history_for_this_job is not None:
                                    old_history_for_this_job = json.loads(
                                        old_history_for_this_job
                                    )
                                else:
                                    old_history_for_this_job = {}
                                if job_failed_last_time:
                                    ljt(
                                        "Job failed last time (new jobs were generated, graph rerun, not running again: {job_id}"
                                    )
                                    outputs = None
                                    error = self.last_job_states[job_id].payload
                                else:
                                    outputs = job.run(
                                        self, old_history_for_this_job
                                    )  # job_state.historical_output)
                                # ljt(f"job successfull {job_id}")
                            finally:
                                # we still check the cwd, even if the job failed!
                                if os.getcwd() != cwd:
                                    os.chdir(
                                        cwd
                                    )  # restore and hope we can recover enough to actually print the exception, I suppose.
                                    log_error(
                                        f"{job_id} changed current_working_directory. Since ppg2 is multithreaded, you must not do this in jobs that RunHere"
                                    )
                                    error = exceptions.JobContractError(
                                        f"{job_id} changed current_working_directory. Since ppg2 is multithreaded, you must not do this in jobs that RunHere"
                                    )
                                    outputs = None
                                    raise error

                    except (
                        SystemExit
                    ) as e:  # pragma: no cover - happens in spawned process, and we don't get coverage logging for it thanks to os._exit
                        log_trace(
                            "SystemExit in spawned process -> converting to hard exit"
                        )
                        if os.getpid() != self.pid:
                            os._exit(e.args[0])
                    except Exception as e:
                        if isinstance(e, KeyboardInterrupt):  # happens on abort
                            raise
                        elif isinstance(e, exceptions.JobError):
                            pass  # take it at face value
                        else:
                            exception_type, exception_value, tb = sys.exc_info()
                            captured_tb = ppg_traceback.Trace(
                                exception_type, exception_value, tb
                            )
                            print(captured_tb)
                            e = exceptions.JobError(
                                e,
                                captured_tb,
                            )
                        error = e
                    finally:
                        if job_id is not None:
                            job.stop_time = time.time()
                            job.run_time = job.stop_time - job.start_time
                            # ljt(f"end {job_id} {self.jobs_in_flight}")
                            # log_trace(f"Leaving thread for {job_id}")
                            with self.evaluator_lock:
                                if outputs is None and error is None:
                                    # this happes when the job get's terminated by abort
                                    log_error(f"job aborted {job_id}")
                                    error = exceptions.JobError(
                                        "Aborted", KeyboardInterrupt()
                                    )
                                    # raise ValueError("Should not happen")
                                if outputs is not None and set(outputs.keys()) != set(
                                    self.jobs[job_id].outputs
                                ):  # the 2nd one is a list...
                                    log_trace(
                                        f"\t{job_id} returned the wrong set of outputs. "
                                        f"Should be {escape_logging(str(set(self.jobs[job_id].outputs)))}, was {escape_logging(str(set(outputs.keys())))}"
                                    )

                                    error = exceptions.JobContractError(
                                        f"\t{job_id} returned the wrong set of outputs. "
                                        f"Should be {escape_logging(str(set((self.jobs[job_id].outputs))))}, was {escape_logging(str(set(outputs.keys())))}",
                                    )
                                    outputs = None

                                if outputs is not None:
                                    self.job_outcomes[job_id] = RecordedJobOutcome(
                                        job_id, JobOutcome.Success, outputs
                                    )
                                    str_history = json.dumps(
                                        outputs, sort_keys=True, indent=1
                                    )
                                    ljt(f"success {job_id} str_history {str_history}")
                                    try:
                                        self.evaluator.event_job_success(
                                            job_id, str_history
                                        )
                                        failed = False
                                    except Exception as e:
                                        log_error(
                                            f"Recording job success failed for {job_id}. Likely constraint violation?: Message was '{e}'"
                                        )
                                        self.job_outcomes[job_id] = RecordedJobOutcome(
                                            job_id, JobOutcome.Failed, str(e)
                                        )

                                else:
                                    ljt(f"failure {job_id} - {error}")
                                    self.job_outcomes[job_id] = RecordedJobOutcome(
                                        job_id, JobOutcome.Failed, error
                                    )
                                    self.evaluator.event_job_failure(job_id)
                                    try:
                                        self.log_failed_job(job_id, error)
                                    except Exception as e:
                                        log_error(
                                            f"logging job failure failed {job_id} {e}"
                                        )
                                self.jobs_in_flight.remove(job_id)
                                if c > 1:
                                    self.jobs_all_cores_in_flight -= 1

                            self.job_outcomes[job_id].run_time = job.run_time
                            self.check_for_new_jobs.set()
                        elif error:
                            print(error)
                            self.check_for_new_jobs.set()
                            raise error

            except (KeyboardInterrupt, SystemExit):  # happens on abort
                log_trace(
                    f"Keyboard Interrupt received {time.time() - self.abort_time}"
                )
                pass
            except Exception as e:
                log_error(
                    f"Captured exception outside of loop - should not happen {type(e)} {str(e)}. Check error log. Run is being aborted"
                )
                exception_type, exception_value, tb = sys.exc_info()
                captured_tb = ppg_traceback.Trace(exception_type, exception_value, tb)
                log_error(str(captured_tb))
                self.abort()
                raise
            except BaseException as e:
                log_error(f"panic from the rust side {e}. Killing process")
                os.kill(os.getpid(), signal.SIGTERM)
                sys.exit(1)
        finally:
            # log_job_trace(f"left thread {len(self.threads)} {job_id}")
            ...


class JobCollector:
    """only in place during the dag modification step of Runner.__init__,
    so that the jobs that are only created during run (cleanup, )
    do not end up in the actual graph.
    """

    def __init__(self, run_mode):
        self.clear()
        self.run_mode = run_mode

    def add(self, job):
        self.jobs[job] = job

    def clear(self):
        self.jobs = {}
        self.edges = set()



from pathlib import Path
import pytest
import pypipegraph2 as ppg
from .shared import write, read, append, writeappend, Dummy, counter

shu = None


@pytest.mark.usefixtures("create_out_dir")
@pytest.mark.usefixtures("ppg2_per_test")
class TestJobGeneratingJob:
    def test_basic(self):
        def gen():
            ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "A"))
            ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))
            ppg.FileGeneratingJob("out/C", lambda of: write("out/C", "C"))

        ppg.JobGeneratingJob("genjob", gen)
        ppg.run()
        assert read("out/A") == "A"
        assert read("out/B") == "B"
        assert read("out/C") == "C"

    def test_injecting_multiple_stages(self):
        def gen():
            def genB():
                def genC():
                    ppg.FileGeneratingJob("out/D", lambda of: write("out/D", "D"))

                ppg.JobGeneratingJob("C", genC)

            ppg.JobGeneratingJob("B", genB)

        ppg.JobGeneratingJob("A", gen)
        ppg.run()
        assert read("out/D") == "D"

    def test_generated_job_depending_on_each_other_one_of_them_is_Invariant(self):
        # basic idea. You have jobgen A,
        # it not only creates filegenB, but also ParameterDependencyC that A depends on
        # does that work
        def gen():
            jobB = ppg.FileGeneratingJob(
                "out/B", lambda of: write("out/B", "B"), depend_on_function=False
            )
            jobC = ppg.ParameterInvariant("C", ("ccc",))
            jobB.depends_on(jobC)

        ppg.JobGeneratingJob("A", gen)
        ppg.run()
        assert read("out/B") == "B"

        ppg.new()

        def gen2():
            jobB = ppg.FileGeneratingJob(
                "out/B", lambda of: write("out/B", "C"), depend_on_function=False
            )
            jobC = ppg.ParameterInvariant("C", ("ccc",))
            jobB.depends_on(jobC)

        ppg.JobGeneratingJob("A", gen2)
        ppg.run()
        assert read("out/B") == "B"  # no rerun

        ppg.new()

        def gen3():
            counter("3")
            jobB = ppg.FileGeneratingJob(
                "out/B", lambda of: write("out/B", "C"), depend_on_function=False
            )
            # jobB.ignore_code_changes()
            jobCX = ppg.ParameterInvariant("C", ("DDD",))
            jobB.depends_on(jobCX)

        ppg.JobGeneratingJob("A", gen3)
        ppg.run()
        assert read("out/B") == "C"  # did get rerun
        assert read("3") == "1"  # check that gen3 really ran...

    def test_generated_job_depending_on_job_that_cant_have_finished(self):
        # basic idea. You have jobgen A, and filegen B.
        # filegenB depends on jobgenA.
        # jobGenA created C depends on filegenB
        # Perhaps add a filegen D that's independand of jobGenA, but C also deps on D
        def a():
            jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))

            def genA():
                jobC = ppg.FileGeneratingJob("out/C", lambda of: write("out/C", "C"))
                jobC.depends_on(jobB)

            jobA = ppg.JobGeneratingJob("A", genA)
            jobB.depends_on(jobA)
            ppg.run()
            assert read("out/B") == "B"
            assert read("out/C") == "C"

        def b():
            jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))
            jobD = ppg.FileGeneratingJob("out/D", lambda of: write("out/D", "D"))

            def genA():
                jobC = ppg.FileGeneratingJob("out/C", lambda of: write("out/C", "C"))
                jobC.depends_on(jobB)
                jobC.depends_on(jobD)

            jobA = ppg.JobGeneratingJob("A", genA)
            jobB.depends_on(jobA)
            ppg.run()
            assert read("out/B") == "B"
            assert read("out/C") == "C"

        a()
        ppg.new()
        b()

    def test_generated_job_depending_on_each_other(self):
        # basic idea. You have jobgen A,
        # it not only creates filegenB, but also filegenC that depends on B
        # does that work
        def gen():
            jobB = ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "B"))
            jobC = ppg.FileGeneratingJob(
                "out/C", lambda of: write("out/C", read("out/B"))
            )
            jobC.depends_on(jobB)

        ppg.JobGeneratingJob("A", gen)
        ppg.run()
        assert read("out/B") == "B"
        assert read("out/C") == "B"

    def test_generated_job_depending_on_each_other_one_of_them_is_loading(self):
        # basic idea. You have jobgen A,
        # it not only creates filegenB, but also DataloadingC that depends on B
        # does that work
        def gen():
            def load():
                global shu
                shu = "123"

            def do_write(of):
                global shu
                write(of, shu)

            dl = ppg.DataLoadingJob("dl", load)
            jobB = ppg.FileGeneratingJob("out/A", do_write)
            jobB.depends_on(dl)

        ppg.JobGeneratingJob("gen", gen)
        ppg.run()
        assert read("out/A") == "123"

    def test_passing_non_function(self):
        with pytest.raises(TypeError):
            ppg.JobGeneratingJob("out/a", "shu")

    def test_passing_non_string_as_jobid(self):
        with pytest.raises(TypeError):
            ppg.JobGeneratingJob(5, lambda: 1)

    def test_generated_jobs_that_can_not_run_right_away_because_of_dataloading_do_not_crash(
        self,
    ):
        o = Dummy()
        existing_dl = ppg.AttributeLoadingJob("a", o, "a", lambda: "Ashu")

        def gen():
            new_dl = ppg.AttributeLoadingJob("b", o, "b", lambda: "Bshu")
            fg_a = ppg.FileGeneratingJob("out/C", lambda of: write("out/C", o.a))
            fg_b = ppg.FileGeneratingJob("out/D", lambda of: write("out/D", o.b))
            fg_a.depends_on(existing_dl)
            fg_b.depends_on(new_dl)

        ppg.JobGeneratingJob("E", gen)
        ppg.run()
        assert read("out/C") == "Ashu"
        assert read("out/D") == "Bshu"

    def test_filegen_invalidated_jobgen_created_filegen_later_also_invalidated(self):
        a = ppg.FileGeneratingJob(
            "out/A", lambda of: writeappend("out/A", "out/Ac", "A")
        )
        p = ppg.ParameterInvariant("p", "p")
        a.depends_on(p)

        def gen():
            append("out/g", "g")
            c = ppg.FileGeneratingJob(
                "out/C", lambda of: writeappend("out/C", "out/Cx", "C")
            )
            c.depends_on(a)

        # difference to ppg1 - gen jobs will not rerun if their inputs did not change!
        ppg.JobGeneratingJob("b", gen).depends_on(a)
        ppg.run()
        assert read("out/g") == "g"
        assert read("out/A") == "A"
        assert read("out/Ac") == "A"
        assert read("out/C") == "C"
        assert read("out/Cx") == "C"

        ppg.new()
        a = ppg.FileGeneratingJob(
            "out/A", lambda of: writeappend("out/A", "out/Ac", "A")
        )
        p = ppg.ParameterInvariant("p", "p2")
        a.depends_on(p)
        ppg.JobGeneratingJob("b", gen).depends_on(a)
        ppg.run()
        assert read("out/g") == "gg"
        assert read("out/Ac") == "AA"
        assert (
            read("out/Cx") == "C"
        )  # this is a difference to the pipegraph. depending on A
        # does not retrigger just because a's upstream changed.
        # only if a truly changed

        ppg.new()
        a = ppg.FileGeneratingJob(
            "out/A", lambda of: writeappend("out/A", "out/Ac", "B")
        )
        p = ppg.ParameterInvariant("p", "p3")
        a.depends_on(p)
        ppg.JobGeneratingJob("b", gen).depends_on(a)
        ppg.run()
        assert read("out/g") == "ggg"
        assert read("out/Ac") == "AAB"
        assert read("out/Cx") == "CC"

    def test_creating_within_dataload(self):
        """This used to be forbidden in ppg1.
        I don't see a reason to forbid it now, the only
        substantial difference is that DataLoadingJobs run when their downstream needs them,
        and JobGeneratingJobs always run"""
        write_job = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "aa"))

        def load():
            ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "aa"))

        dl = ppg.DataLoadingJob("load_data", load)
        write_job.depends_on(dl)
        # with pytest.raises(ppg.JobsFailed):
        ppg.run()
        assert Path("out/B").exists()

    def test_ignored_if_generating_within_filegenerating(self):
        write_job = ppg.FileGeneratingJob("out/A", lambda of: write("out/A", "aa"))

        def load(_of):
            ppg.FileGeneratingJob("out/B", lambda of: write("out/B", "aa"))
            write("out/C", "c")

        dl = ppg.FileGeneratingJob("out/C", load)
        write_job.depends_on(dl)
        ppg.run()
        assert read("out/C") == "c"
        assert not Path("out/B").exists()

    def test_invalidation(self):
        def gen():
            ppg.FileGeneratingJob("out/D", lambda of: write("out/D", "D"))

        ppg.JobGeneratingJob("A", gen)
        ppg.run()
        assert read("out/D") == "D"
        ppg.new()

        def gen():
            ppg.FileGeneratingJob("out/D", lambda of: write("out/D", "E"))

        ppg.JobGeneratingJob("A", gen)
        ppg.run()
        assert read("out/D") == "E"

    def test_invalidation_multiple_stages(self):
        counter = [0]

        def count():
            counter[0] += 1
            return str(counter[0])

        def gen():
            def genB():
                def genC():
                    count()
                    ppg.FileGeneratingJob("out/D", lambda of: write(of, "D"))

                ppg.JobGeneratingJob("C", genC)

            ppg.JobGeneratingJob("B", genB)

        ppg.JobGeneratingJob("A", gen)
        ppg.run()
        assert read("out/D") == "D"
        assert counter[0] == 1

        ppg.new()
        ppg.JobGeneratingJob("A", gen)
        ppg.run()
        assert read("out/D") == "D"
        assert counter[0] == 2

        ppg.new()

        def gen():
            def genB():
                def genC():
                    count()
                    ppg.FileGeneratingJob("out/D", lambda of: write("out/D", "E"))

                ppg.JobGeneratingJob("C", genC)

            ppg.JobGeneratingJob("B", genB)

        ppg.JobGeneratingJob("A", gen)
        ppg.run()
        assert read("out/D") == "E"
        assert counter[0] == 3

    # this test only works if you'd remove the locking around next_job_number
    # *and* add in a small delay to actually trigger it
    # def test_massive_trying_to_hit_jobnumber_conflict(self):
    #     ppg.new(cores=4)
    #     def fail(of):
    #         raise ValueError()
    #     def jj(prefix):
    #         def inner():
    #             Path(prefix).mkdir(exist_ok=True)
    #             for i in range(0,100):
    #                 ppg.FileGeneratingJob(prefix + '/' + str(i), lambda of: of.write_text(prefix))
    #         return inner
    #     a = ppg.JobGeneratingJob('a', jj('a'))
    #     b = ppg.JobGeneratingJob('b', jj('b'))
    #     c = ppg.JobGeneratingJob('c', jj('c'))
    #     d = ppg.JobGeneratingJob('d', jj('d'))
    #     ppg.run()
